/*******************************************************************************
* raw2ts.cpp: raw -> ES ffmpeg encoder -> TS
*-------------------------------------------------------------------------------
* (c)1999-2001 VideoLAN
* $Id: raw2ts.cpp,v 1.25 2003/03/20 12:30:46 titer Exp $
*
* Authors: Tristan Leteurtre <tooney@videolan.org>
*          Eric Petit <titer@videolan.org>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*-------------------------------------------------------------------------------
*
*******************************************************************************/


//------------------------------------------------------------------------------
// Preamble
//------------------------------------------------------------------------------
#include "../../core/defs.h"

#include "../../core/core.h"

#include "../../mpeg/mpeg.h"
#include "../../mpeg/ts.h"
#include "../../mpeg/pes.h"

#include "../../server/buffer.h"
#include "../../server/program.h"
#include "../../server/broadcast.h"
#include "../../server/request.h"
#include "../../server/input.h"

#include "../../mpeg/reader.h"
#include "../../mpeg/converter.h"

#ifdef HAVE_DVBPSI_DVBPSI_H
#   include <dvbpsi/dvbpsi.h>
#   include <dvbpsi/psi.h>
#   include <dvbpsi/descriptor.h>
#   include <dvbpsi/pat.h>
#   include <dvbpsi/pmt.h>
#else
#   include "src/dvbpsi.h"
#   include "src/psi.h"
#   include "src/descriptor.h"
#   include "src/tables/pat.h"
#   include "src/tables/pmt.h"
#endif

#include "../../mpeg/dvbpsi.h"

#include "raw2ts.h"

#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/mman.h>
#include <unistd.h>

/* For example : 0,1s = 100000us -> about 10000 in 90 kHz clock */
#define DELTA_PCR_PTS 10000


/* Usefull */
u64 GetTime(void)
{
  struct timeval tmp;
  gettimeofday(&tmp, NULL);
  return( (u64)tmp.tv_sec * 1000000 + (u64)tmp.tv_usec );
}


//------------------------------------------------------------------------------
// Library declaration
//------------------------------------------------------------------------------
#ifdef __PLUGIN__
GENERATE_LIB_ARGS(C_Raw2TsMpegConverterModule, handle);
#endif


//------------------------------------------------------------------------------
// Builtin declaration
//------------------------------------------------------------------------------
#ifdef __BUILTIN__
C_Module* NewBuiltin_raw2ts(handle hLog)
{
  return new C_Raw2TsMpegConverterModule(hLog);
}
#endif


/*******************************************************************************
* C_Raw2TsConverter
*******************************************************************************/

//------------------------------------------------------------------------------
// Constructor
//------------------------------------------------------------------------------
C_Raw2TsMpegConverter::C_Raw2TsMpegConverter(C_Module* pModule,
                                           C_MpegConverterConfig& cConfig) :
                            C_MpegConverter(pModule, cConfig)
{
  m_VideoCodec = NULL;
  m_VideoCodecContext = NULL;
  m_AudioCodec = NULL;
  m_AudioCodecContext = NULL;
}


//------------------------------------------------------------------------------
// Initialization
//------------------------------------------------------------------------------
void C_Raw2TsMpegConverter::InitWork()
{
  // Inherited method
  C_MpegConverter::InitWork();

  iVideoContinuityCounter = 0;
  iAudioContinuityCounter = 0;

  // get configuration
  m_iHeight = m_pBroadcast->GetOption("FrameHeight").ToInt();
  m_iWidth = m_pBroadcast->GetOption("FrameWidth").ToInt();
  m_iFrameRate = m_pBroadcast->GetOption("FrameRate").ToInt();
  m_iPalette = m_pBroadcast->GetOption("Palette").ToInt();
  m_bDeInterlace = m_pBroadcast->GetOption("v4l_deinterlace").ToInt();
  m_strCompression = m_pBroadcast->GetOption("v4l_codec") ;
  m_audioCompression = m_pBroadcast->GetOption("v4l_audiocodec") ;
  m_channel = m_pBroadcast->GetOption("v4l_audiochannel").ToInt() ;
  m_audiofreq = m_pBroadcast->GetOption("v4l_audiofreq").ToInt() ;
  m_quality = atof(m_pBroadcast->GetOption("v4l_quality").GetString()) ;

  if (m_bDeInterlace)
    fprintf( stderr, "Info : Image de-interlacement ON\n");

  // Init ffmpeg AV codec stuff
  avcodec_init();
  avcodec_register_all();

  if(m_strCompression == "mpeg4") {
    m_type = 0x10 ;
  }
  else if(m_strCompression == "mpeg1video") {
    m_type = 0x01 ; /* iso 11172 video */
  }
  else { // All other encoders are not working with vlc.
    m_type = 0x01 ; /* iso 11172 video */
    m_strCompression = "mpeg1video" ;
  }
  // prepare the video encoder
  m_VideoCodec = avcodec_find_encoder_by_name(m_strCompression.GetString());
  if (!m_VideoCodec)
  {
    m_strCompression="mpeg1video";
    m_VideoCodec = avcodec_find_encoder(CODEC_ID_MPEG1VIDEO);
  }
  if (!m_VideoCodec)
  {
    throw E_Exception(GEN_ERR,
                      "Cannot find encoder " + m_strCompression + " Video");
  }

  // prepare the audio encoder
  // register_avcodec(&mp2_encoder);
  m_AudioCodec = avcodec_find_encoder_by_name( m_audioCompression.GetString() );
  if( !m_AudioCodec )
  {
    m_audioCompression="mp2" ;
    m_AudioCodec = avcodec_find_encoder(CODEC_ID_MP2);
  }
  if( !m_AudioCodec )
  {
    throw E_Exception(GEN_ERR, "Cannot find encoder " + m_audioCompression + " Audio");
  }


  // video parameters
  m_VideoCodecContext = avcodec_alloc_context();
  m_VideoCodecContext->bit_rate =
    m_pBroadcast->GetOption("v4l_bitrate").ToInt() * 1000;
  m_VideoCodecContext->bit_rate_tolerance =
    m_VideoCodecContext->bit_rate / 5; // 20%
  m_VideoCodecContext->width = m_iWidth;
  m_VideoCodecContext->height = m_iHeight;
  m_VideoCodecContext->frame_rate = m_iFrameRate;
  m_VideoCodecContext->frame_rate_base = 1;
  m_VideoCodecContext->gop_size =
    m_iFrameRate; // emit one intra frame every second
  m_VideoCodecContext->pix_fmt = (enum PixelFormat)m_iPalette;
  m_VideoCodecContext->qmin = 2;
  m_VideoCodecContext->qmax = 31;
  m_VideoCodecContext->max_qdiff = 5;
  //m_VideoCodecContext->qblur = 0.5;
  //m_VideoCodecContext->qcompress = 0.5;
  //m_VideoCodecContext->me_method = 0;

  switch(m_iPalette)
  {
    case PIX_FMT_YUV420P:
      m_iPictureBufSize = (m_iHeight * m_iWidth) * 3 / 2 ;
      break;
    case PIX_FMT_YUV422:
      m_iPictureBufSize = (m_iHeight * m_iWidth) * 2 ;
      break;
    case PIX_FMT_BGR24:
      m_iPictureBufSize = (m_iHeight * m_iWidth) * 3 ;
      break;
    default:
      fprintf( stderr, "Warning : unknown palette\n" );
      m_iPictureBufSize = 0;
      break;
  }

  // audio parameters
  m_AudioCodecContext = avcodec_alloc_context();
  m_AudioCodecContext->sample_rate = m_audiofreq ;
  m_AudioCodecContext->channels = m_channel ;

  m_AudioCodecContext->bit_rate =
    m_pBroadcast->GetOption("v4l_audiobitrate").ToInt() * 1000;
/* Only in mp2 stereo !!
  if( m_AudioCodecContext->bit_rate < 32 )
  {
    // ffmpeg does not like bitrate under 32000, back to default
    m_AudioCodecContext->bit_rate = 64;
  }
*/

  // "open" the video encoder
  if (avcodec_open(m_VideoCodecContext, m_VideoCodec) < 0)
  {
    throw E_Exception(GEN_ERR, "Cannot init video encoder");
  }
  /* Be carefull ! Some parameters may have changed, like palette ! */

  fprintf(stderr, "Info : Encoding : %ux%u %u fps at %u kb/s in %s\n",
                  m_iWidth,m_iHeight,m_iFrameRate,
                  m_VideoCodecContext->bit_rate / 1000,
                  m_strCompression.GetString());

  // "open" the audio encoder
  if( avcodec_open( m_AudioCodecContext, m_AudioCodec ) < 0 )
  {
    throw E_Exception(GEN_ERR, "Cannot init audio encoder");
  }

  // prepare buffers
  m_iVideoOutBufSize = 500000;
  m_VideoOutBuf = new u8 [m_iVideoOutBufSize];
  m_Picture = (AVPicture *)malloc(sizeof(AVPicture));

  m_iAudioSampleBufSize = 2 * m_AudioCodecContext->frame_size *
                          m_AudioCodecContext->channels;
  m_iAudioOutBufSize = 10000; // FIXME: totally arbitrary !
  m_AudioOutBuf = (byte*)malloc( m_iAudioOutBufSize );
}


//------------------------------------------------------------------------------
// Main job
//------------------------------------------------------------------------------
void C_Raw2TsMpegConverter::DoWork()
{
  // Initialize the condition in the new thread.
  m_cResumeCond.Protect();

  C_String strPgrmName = m_pBroadcast->GetProgram()->GetName();
  LogDbg(m_hLog, "Starting to read program \"" + strPgrmName + "\"");

  // start recording to fifos (size = 1 sec)
  C_Fifo<C_DatedBuffer> * pVideoFifo =
    new C_Fifo<C_DatedBuffer>( m_iFrameRate, YES, true );
  C_Fifo<C_DatedBuffer> * pAudioFifo =
    new C_Fifo<C_DatedBuffer>( 2 * m_channel * m_audiofreq /
                               m_iAudioSampleBufSize, YES, true );

  C_Condition * pHasDataCondition = new C_Condition();
  C_Mutex * pVideoMutex = new C_Mutex();
  C_Mutex * pAudioMutex = new C_Mutex();
  C_VideoCaptureThread * videoThread =
    new C_VideoCaptureThread( m_pReader, pVideoFifo,
                              m_iPictureBufSize, pVideoMutex,
                              pHasDataCondition, m_quality );
//  C_AudioCaptureThread * audioThread =
//    new C_AudioCaptureThread( m_pReader, pAudioFifo,
//                              m_iAudioSampleBufSize, pAudioMutex,
//                              pHasDataCondition, m_channel, m_audiofreq );
  videoThread->Create();
//  audioThread->Create();

  /* Build a PAT */
  C_DvbPsiPat * cPat = new C_DvbPsiPat(0, 0, true);
  dvbpsi_PATAddProgram(cPat->GetLowLevelPat(), 0x1212, 0x12);
  cPat->Generate();
  cPat->TsReset();

  /* Build a PMT */
  C_DvbPsiPmt * cPmt = new C_DvbPsiPmt(0x1212, 0, true, 0xE0, 0x12);
  dvbpsi_PMTAddES(cPmt->GetLowLevelPmt(), m_type, 0xE0);

  // XXX TODO : with mpeg4 some descriptors have to be specified in PAT !

  dvbpsi_PMTAddES(cPmt->GetLowLevelPmt(), 3, 0xC0);
  cPmt->Generate();
  cPmt->TsReset();

  // main loop: encode and stream
  int iRc = 0;
  bool bDiscontinuity = false;
  AVPicture * TmpPicture;
  AVFrame m_Frame;
  u8 * TmpBuf;
  C_TsPacket * pPacket;

  while(!m_bStop && !iRc)
  {
    // Suspend/Resume handling
    if(m_iShortPauseStatus == STATUS_PAUSE_REQUEST)
    {
      m_cResumeCond.Wait();
      m_iShortPauseStatus = STATUS_RUNNING;
    }

    if(m_iLongPauseStatus == STATUS_PAUSE_REQUEST)
    {
      m_cResumeCond.Wait();
      m_iLongPauseStatus = STATUS_RUNNING;
      bDiscontinuity = true;
    }

    // DO THE REAL WORK

    // if both fifos are empty, wait for a frame
    if( !pAudioFifo->Size() && !pVideoFifo->Size() )
    {
      pHasDataCondition->Protect();
      pHasDataCondition->Wait();
      pHasDataCondition->Release();
    }

    // encoding audio is our priority if the CPU is too slow;
    // otherwise, encode data from the most filled fifo
    C_DatedBuffer * datedBuffer;


/* 

 if( pAudioFifo->Size() > pAudioFifo->Capacity() / 2 ||
        100 * pAudioFifo->Size() / pAudioFifo->Capacity() >
          100 * pVideoFifo->Size() / pVideoFifo->Capacity() )
    {
      int outSize;

      pAudioMutex->Lock();
*/
      datedBuffer = pAudioFifo->Pop();
/*
      pAudioMutex->UnLock();

      // encode the buffer
      outSize = avcodec_encode_audio( m_AudioCodecContext, m_AudioOutBuf,
                                      datedBuffer->m_iSize,
                                      (short*)datedBuffer->m_pData );

      // build TS packets and send them
      EStoTS( m_AudioOutBuf, outSize, datedBuffer->m_TimeStamp,
              0xC0, &iAudioContinuityCounter );

*/

      delete datedBuffer;
/*

    }
    else
    {

*/

      pVideoMutex->Lock();
      datedBuffer = pVideoFifo->Pop();
      pVideoMutex->UnLock();
      if( datedBuffer != NULL )
      {
        byte * to_free = NULL; // for deinterlacing

        avpicture_fill(m_Picture, datedBuffer->m_pData,
                       m_VideoCodecContext->pix_fmt,
                       m_iWidth, m_iHeight);

        if (m_bDeInterlace)
          DeInterlace(m_VideoCodecContext, m_Picture, &to_free );

        // Palette convertions, if needed
        if (m_iPalette != m_VideoCodecContext->pix_fmt)
        {
          avpicture_fill(TmpPicture, TmpBuf, m_VideoCodecContext->pix_fmt,
                         m_VideoCodecContext->width,
                         m_VideoCodecContext->height);
          if (img_convert(TmpPicture, m_VideoCodecContext->pix_fmt,
                          m_Picture, m_iPalette,
                          m_VideoCodecContext->width,
                          m_VideoCodecContext->height) < 0)
          {
            fprintf(stderr, "Could not convert image\n");
          }
          m_Picture = TmpPicture;
        }

        *(AVPicture*)&m_Frame= *m_Picture;

        m_Frame.quality = m_quality;

        // encode the video frame
        m_iVideoOutSize = avcodec_encode_video(m_VideoCodecContext,
                                               m_VideoOutBuf,
                                               m_iVideoOutBufSize, &m_Frame);
        // stream it
        EStoTS( m_VideoOutBuf, m_iVideoOutSize, datedBuffer->m_TimeStamp,
                0xE0, &iVideoContinuityCounter );

        if( to_free ) free( to_free ); // free the deinterlace picture
        delete datedBuffer; // free the original picture
      }


 // }  




    // send some PAT and PMT !
    // kludge : Send sometime, not too often, not too rare...
    // ContinuityCounter <= 2 is an arbitrary criteria

    if (iVideoContinuityCounter <= 2)
    {
      pPacket = m_pTsProvider->GetPacket();
      cPat->TsReset();
      cPat->TsWrite(pPacket);
      m_pHandler->HandlePacket(pPacket);

      pPacket = m_pTsProvider->GetPacket();
      cPmt->TsReset();
      cPmt->TsWrite(pPacket);
      m_pHandler->HandlePacket(pPacket);
    }

    if(iRc == MPEG_ENDOFSTREAM)
    {
      C_String strPgrmName = m_pBroadcast->GetProgram()->GetName();
      LogDbg(m_hLog, "End of program \"" + strPgrmName + "\" reached");
    }
    else if (iRc)
    {
      C_String strPgrmName = m_pBroadcast->GetProgram()->GetName();
      Log(m_hLog, LOG_ERROR, "Read error for program \"" + strPgrmName + "\"");
    }


  }

  if(!m_bStop)
  {
    LogDbg(m_hLog,
           "Stopping converter by callback for program " + strPgrmName);
    m_bStop = true;
    C_Event cEvent(m_pBroadcast->GetInput()->GetName());
    cEvent.SetCode(EOF_EVENT);
    cEvent.SetBroadcast(m_pBroadcast);
    m_pEventHandler->HandleEvent(cEvent);
  }

  m_cResumeCond.Release();

  LogDbg(m_hLog, "Converter stopped for program " + strPgrmName);
}

// Packetizer :
//    - ADDs PES Header
//    - packtizes in TS packets
void C_Raw2TsMpegConverter::EStoTS( byte * EsData, int EsSize, int TimeStamp,
                                    u8 iPayloadType, int * iCC)
{
  C_PesPacket * pPesPacket = new C_PesPacket();
  int iTmpEsSize;
  bool bFirstPES = true;

  for( iTmpEsSize = EsSize; iTmpEsSize > 0;
       iTmpEsSize -= ( PES_MAX_SIZE - 20 ) )
  {
    // maximum ( PES_MAX_SIZE - 20 ) of ES data because we must let
    // some space for PES headers
    pPesPacket->BuildPacket( bFirstPES ? TimeStamp : 0, 0xE0,
                             EsData + EsSize - iTmpEsSize,
                             iTmpEsSize > ( PES_MAX_SIZE - 20 ) ?
                               ( PES_MAX_SIZE - 20 ) : iTmpEsSize);
    bFirstPES = false;

    C_TsPacket * pTsPacket;
    unsigned int iTmpPesSize, iHeader;
    bool bUnitStart = true;

    for( iTmpPesSize = pPesPacket->GetSize(); iTmpPesSize > 0;
         iTmpPesSize -= (TS_PACKET_LEN - iHeader ) )
    {
      // add TS headers
      pTsPacket = m_pTsProvider->GetPacket();
      pTsPacket->BuildHeader(iPayloadType, bUnitStart, (*iCC)++);
      iHeader = 4;

      if (bUnitStart)
      {
        pTsPacket->BuildAdaptionField((u64)(TimeStamp - (u64)DELTA_PCR_PTS));
        iHeader += 8;
        bUnitStart = false;
      }

      if (iTmpPesSize < TS_PACKET_LEN - iHeader)
        iHeader = pTsPacket->AddStuffingBytes(iTmpPesSize);

      pTsPacket->FillWith(*pPesPacket + pPesPacket->GetSize() - iTmpPesSize,
                        iHeader  , TS_PACKET_LEN - iHeader);

      m_pHandler->HandlePacket(pTsPacket);
    }
    (*iCC) &= 0xF;
  }

  delete pPesPacket;
}

void C_Raw2TsMpegConverter::DeInterlace(AVCodecContext *pCodec,
                                        AVPicture *pPicture, byte ** to_free)
{
  /* See ffmpeg.c */
  u8 * pBuf = NULL;
  AVPicture pTmpPicture;
  AVPicture * pNewPicture = &pTmpPicture;

  int size;

  /* create temporary picture */
  size = avpicture_get_size(pCodec->pix_fmt, pCodec->width, pCodec->height);
  pBuf = (u8*)malloc( size );
  if (!pBuf)
    throw E_Exception(GEN_ERR, "Cannot init deinterlace buffer");

  avpicture_fill(pNewPicture, pBuf, pCodec->pix_fmt,
                 pCodec->width, pCodec->height);

  if (avpicture_deinterlace(pNewPicture, pPicture, pCodec->pix_fmt,
                            pCodec->width, pCodec->height) < 0)
  {
    /* if error, do not deinterlace */
    pNewPicture = pPicture;
  }

  if (pPicture != pNewPicture)
    *pPicture = *pNewPicture;

  *to_free = pBuf;
}


/*******************************************************************************
* C_VideoCaptureThread
*******************************************************************************/

C_VideoCaptureThread::C_VideoCaptureThread( C_MpegReader * pReader,
                                            C_Fifo<C_DatedBuffer> *pFifo,
                                            int iPictureBufSize,
                                            C_Mutex * pMutex,
                                            C_Condition * pCondition,
                                            float quality )
{
  m_pReader = pReader;
  m_pFifo = pFifo;
  m_iPictureBufSize = iPictureBufSize;
  m_pMutex = pMutex;
  m_pCondition = pCondition;
  m_quality = quality ;
}

void C_VideoCaptureThread::InitWork()
{
  m_bStop = false;
}

void C_VideoCaptureThread::DoWork()
{
  C_DatedBuffer * pDatedBuffer;
  while( !m_bStop )
  {
    // allocate the needed memory
    // TODO: use a packet provider to avoid doing a malloc at each loop
    pDatedBuffer = new C_DatedBuffer();
    if( !pDatedBuffer->Alloc( m_iPictureBufSize ) )
    {
      fprintf( stderr, "Error : cannot Alloc() (%d bytes)\n",
                       m_iPictureBufSize );
      break;
    }

    // read data from the device; this is the blocking stage.
    byte * tmp = NULL;
    pDatedBuffer->m_iSize =
      m_pReader->GetFrame( &tmp, m_iPictureBufSize );

    // date it
    pDatedBuffer->m_TimeStamp = ( GetTime() * 9 / 100 ) &  0x0effffff;

    // the memory returned by the reader does not belong to us, so we
    // have to memcpy it before pushing this into the fifo.
    // TODO: find a way to avoid memcpy-ing
    memcpy( pDatedBuffer->m_pData, tmp, m_iPictureBufSize );

    // add it to fifo
    m_pMutex->Lock();
    if( m_pFifo->Size() == m_pFifo->Capacity() )
      fprintf( stderr, "Warning : trashing a video frame (%lld)... "
                       "your computer is too slow !\n",
                       pDatedBuffer->m_TimeStamp );
    m_pFifo->Push( pDatedBuffer );
    m_pMutex->UnLock();

    // warn the encoder thread we have something
    m_pCondition->Protect();
    m_pCondition->Signal();
    m_pCondition->Release();
  }
}

void C_VideoCaptureThread::StopWork()
{
  m_bStop = true;
}

void C_VideoCaptureThread::CleanWork()
{
}


/*******************************************************************************
* C_AudioCaptureThread
*******************************************************************************/

C_AudioCaptureThread::C_AudioCaptureThread( C_MpegReader * pReader,
                                            C_Fifo<C_DatedBuffer> *pFifo,
                                            int iAudioSampleBufSize,
                                            C_Mutex * pMutex,
                                            C_Condition * pCondition,
                                            int channel,
                                            int audiofreq )
{
  m_pReader = pReader;
  m_pFifo = pFifo;
  m_iAudioSampleBufSize = iAudioSampleBufSize;
  m_pMutex = pMutex;
  m_pCondition = pCondition;
  m_channel = channel ;
  m_audiofreq = audiofreq ;
}

void C_AudioCaptureThread::InitWork()
{
  m_bStop = false;
  m_iAudioFD = m_pReader->GetAudioFD();
}



void C_AudioCaptureThread::DoWork()
{
  C_DatedBuffer * pDatedBuffer;
  u64 currentDate;
  struct audio_buf_info buf_info;
  int delay;
// JAVI  while( !m_bStop )

 while( false )
  {
    // allocate the needed memory
    // TODO: use a packet provider to avoid doing a malloc at each loop
    pDatedBuffer = new C_DatedBuffer();
    if( !pDatedBuffer->Alloc( m_iAudioSampleBufSize ) )
    {
      fprintf( stderr, "Error : cannot Alloc() (%d bytes)\n",
                       m_iAudioSampleBufSize );
      break;
    }

    // read data from the device; this is the blocking stage
    delay = read( m_iAudioFD, pDatedBuffer->m_pData, m_iAudioSampleBufSize );

    // date it
    currentDate = GetTime();

    // correct the date because of kernel buffering
    if( ioctl( m_iAudioFD, SNDCTL_DSP_GETISPACE, &buf_info ) == 0 )
      delay += buf_info.bytes;
    currentDate -= (u64)( 1000000 * (float)delay /
                          ( 2.0     /* 16 bits samples */
                            * m_channel     /* channel */
                            * m_audiofreq /* sample rate */ ) );

    pDatedBuffer->m_TimeStamp = ( currentDate * 9 / 100 ) &  0x0effffff;

    // add it to fifo
    m_pMutex->Lock();
    if( m_pFifo->Size() == m_pFifo->Capacity() )
//      fprintf( stderr, "Warning : trashing an audio frame (%lld)... "
//                       "your computer is too slow !\n",
//                       pDatedBuffer->m_TimeStamp );
    m_pFifo->Push( pDatedBuffer );
    m_pMutex->UnLock();

    // warn the encoder thread we have something
    m_pCondition->Protect();
    m_pCondition->Signal();
    m_pCondition->Release();
  }
}

void C_AudioCaptureThread::StopWork()
{
  m_bStop = true;
}

void C_AudioCaptureThread::CleanWork()
{
}

